package eu.jonahbauer.wizard.client.libgdx.screens;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.TextureAtlas;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.scenes.scene2d.Action;
import com.badlogic.gdx.scenes.scene2d.Touchable;
import com.badlogic.gdx.scenes.scene2d.actions.MoveToAction;
import com.badlogic.gdx.scenes.scene2d.ui.Container;
import com.badlogic.gdx.scenes.scene2d.ui.Label;
import com.badlogic.gdx.scenes.scene2d.ui.VerticalGroup;
import com.badlogic.gdx.scenes.scene2d.utils.TextureRegionDrawable;
import com.badlogic.gdx.utils.Align;
import com.badlogic.gdx.utils.I18NBundle;
import eu.jonahbauer.wizard.client.libgdx.AnimationTimings;
import eu.jonahbauer.wizard.client.libgdx.GameAtlas;
import eu.jonahbauer.wizard.client.libgdx.WizardGame;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardActor;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardStack;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardsGroup;
import eu.jonahbauer.wizard.client.libgdx.actors.game.PadOfTruth;
import eu.jonahbauer.wizard.client.libgdx.actors.game.overlay.*;
import eu.jonahbauer.wizard.client.libgdx.state.Game;
import eu.jonahbauer.wizard.client.libgdx.util.Triple;
import eu.jonahbauer.wizard.common.messages.client.InteractionMessage;
import eu.jonahbauer.wizard.common.messages.observer.UserInputMessage;
import eu.jonahbauer.wizard.common.messages.player.ContinueMessage;
import eu.jonahbauer.wizard.common.messages.player.JuggleMessage;
import eu.jonahbauer.wizard.common.messages.player.PlayCardMessage;
import eu.jonahbauer.wizard.common.messages.player.PlayerMessage;
import eu.jonahbauer.wizard.common.model.Card;
import lombok.Getter;
import org.jetbrains.annotations.Nls;

import java.util.*;
import java.util.concurrent.atomic.AtomicBoolean;

import static eu.jonahbauer.wizard.client.libgdx.actions.MyActions.*;
import static eu.jonahbauer.wizard.common.messages.observer.UserInputMessage.Action.*;

public class GameScreen extends MenuScreen {
    @Getter
    private TextureAtlas atlas;

    private Label.LabelStyle labelStyleDefault;
    private Label.LabelStyle labelStyleActive;

    private final Game state;

    private final List<UUID> players;

    private Triple<UUID, UserInputMessage.Action, Long> activePlayer;

    private CardsGroup handCards;
    private CardStack cardStack;
    private PadOfTruth padOfTruth;

    @Getter
    private CardActor trumpCardActor;
    @Getter
    private CardActor trumpSuitActor;

    private VerticalGroup messages;
    private Label persistentMessage;

    private Action currentAction;
    private final Queue<Action> pendingActions = new LinkedList<>();

    private final Map<UUID, Seat> seats = new HashMap<>();
    private final Map<UUID, Label> nameLabels = new HashMap<>();

    private final AtomicBoolean sending = new AtomicBoolean();
    private final AtomicBoolean pendingSync = new AtomicBoolean();

    private boolean juggling;
    private Card juggledCard;

    public GameScreen(WizardGame game) {
        super(game);
        this.state = (Game) game.getClient().getState();
        this.players = new ArrayList<>(state.getPlayers().keySet());
    }

    @Override
    public void show() {
        super.show();
        atlas = new TextureAtlas(Gdx.files.internal(GameAtlas.$PATH));
//        overlay = new TextureRegionDrawable(game.data.uiskinAtlas.findRegion(UiskinAtlas.WHITE)).tint(new Color(0,0,0,0.5f));
        labelStyleDefault = game.data.skin.get(Label.LabelStyle.class);
        labelStyleActive = new Label.LabelStyle(labelStyleDefault);
        labelStyleActive.fontColor = Color.RED;

        seat();
        prepareLabels();

        handCards = new CardsGroup(Collections.emptyList(), atlas);
        handCards.setOnClickListener(this::onCardClicked);
        var container = new Container<>(handCards);
        container.setPosition(360, 75);
        container.setSize(1200, CardActor.PREF_HEIGHT);
        container.fillX();

        messages = new VerticalGroup().columnCenter().bottom();
        messages.setPosition(360, 85 + CardActor.PREF_HEIGHT);
        messages.setSize(1200, 0);
        messages.setTouchable(Touchable.disabled);

        padOfTruth = new PadOfTruth(game.data.skin, new TextureRegionDrawable(atlas.findRegion(GameAtlas.PAD_OF_TRUTH)));
        padOfTruth.setPosition(1910 - 636f, 10);
        padOfTruth.setOrigin(636f, 0);

        cardStack = new CardStack();
        cardStack.setHoverBounds(0.5f * WizardGame.WIDTH - 50, 0.5f * WizardGame.HEIGHT - 50, 100, 100);

        setNames();

        Gdx.input.setInputProcessor(game.data.stage);
        game.data.stage.addActor(container);
        game.data.stage.addActor(cardStack);
        game.data.stage.addActor(padOfTruth);
        game.data.stage.addActor(messages);
    }

    @Override
    protected void renderBackground(float delta) {
        float scale = Math.max(
                game.data.extendViewport.getWorldWidth() / WizardGame.WIDTH,
                game.data.extendViewport.getWorldHeight() / WizardGame.HEIGHT
        );
        game.batch.setColor(1, 1, 1, 0.25f);
        game.batch.draw(game.data.background, 0,0, scale * WizardGame.WIDTH, scale * WizardGame.HEIGHT);
        game.batch.draw(
                game.data.title,
                (game.data.extendViewport.getWorldWidth() - game.data.title.getRegionWidth() * 0.75f) / 2f,
                (game.data.extendViewport.getWorldHeight() - game.data.title.getRegionHeight() * 0.75f) / 2f,
                0.75f * game.data.title.getRegionWidth(),
                0.75f * game.data.title.getRegionHeight()
        );
    }

    @Override
    protected void renderForeground(float delta) {
        if (pendingActions.size() > 0) {
            var actions = game.data.stage.getRoot().getActions();
            boolean running = false;
            if (currentAction != null) {
                for (int i = 0; i < actions.size; i++) {
                    if (actions.get(i) == currentAction) {
                        running = true;
                        break;
                    }
                }
            }

            if (!running) {
                currentAction = pendingActions.poll();
                game.data.stage.addAction(currentAction);
            }
        } else if (pendingSync.getAndSet(false)) {
            send(new ContinueMessage());
        }
    }

    private void execute(Action action) {
        pendingActions.add(action);
    }

    private void execute(Runnable runnable) {
        pendingActions.add(run(runnable));
    }

    private void seat() {
        var count = players.size();
        Seat[] seats = switch (count) {
            case 3 -> new Seat[] {Seat.TOP_LEFT, Seat.TOP_RIGHT};
            case 4 -> new Seat[] {Seat.LEFT, Seat.TOP, Seat.RIGHT};
            case 5 -> new Seat[] {Seat.LEFT, Seat.TOP_LEFT, Seat.TOP_RIGHT, Seat.RIGHT};
            case 6 -> new Seat[] {Seat.LEFT, Seat.TOP_LEFT, Seat.TOP, Seat.TOP_RIGHT, Seat.RIGHT};
            default -> throw new AssertionError();
        };
        int index = players.indexOf(state.getSelf());
        for (int i = 1; i < count; i++) {
            var player = players.get((index + i) % count);
            var seat = seats[i - 1];
            this.seats.put(player, seat);
        }
        this.seats.put(state.getSelf(), Seat.BOTTOM);
    }

    private void prepareLabels() {
        for (UUID player : players) {
            if (isSelf(player)) continue;
            var label = new Label("", game.data.skin);
            var seat = seats.get(player);
            label.setX(seat.getLabelX());
            label.setY(seat.getLabelY());
            label.setAlignment(seat.getLabelAlign());
            this.nameLabels.put(player, label);
            game.data.stage.addActor(label);
        }
    }

    private void setNames() {
        for (int i = 0; i < players.size(); i++) {
            var player = players.get(i);
            var name = state.getPlayers().get(player);
            padOfTruth.setName(i, name);
            if (!isSelf(player)) {
                nameLabels.get(player).setText(name);
            }
        }
    }

    private void onCardClicked(CardActor actor) {
        var card = actor.getCard();

        if (checkAction(PLAY_CARD)) {
            var timeout = activePlayer.third();
            if (card == Card.CLOUD) {
                execute(new PlayColoredCardOverlay(this,
                                                   timeout,
                                                   Card.CLOUD,
                                                   Card.CLOUD_RED,
                                                   Card.CLOUD_GREEN,
                                                   Card.CLOUD_BLUE,
                                                   Card.CLOUD_YELLOW
                ));
            } else if (card == Card.JUGGLER) {
                execute(new PlayColoredCardOverlay(this,
                                                   timeout,
                                                   Card.JUGGLER,
                                                   Card.JUGGLER_RED,
                                                   Card.JUGGLER_GREEN,
                                                   Card.JUGGLER_BLUE,
                                                   Card.JUGGLER_YELLOW
                ));
            } else {
                send(new PlayCardMessage(card));
            }
        } else if (checkAction(JUGGLE_CARD)) {
            juggledCard = card;
            send(new JuggleMessage(card));
        } else {
            addMessage("You cannot do that right now.");
        }
    }

    private boolean checkAction(UserInputMessage.Action action) {
        return activePlayer != null && (activePlayer.first() == null || isSelf(activePlayer.first())) && activePlayer.second() == action;
    }

    private boolean isSelf(UUID uuid) {
        return state.getSelf().equals(uuid);
    }

    public void startRound(int round) {
        execute(parallel(
                run(() -> {
                    if (trumpCardActor != null) {
                        trumpCardActor.remove();
                    }

                    if (trumpSuitActor != null) {
                        trumpSuitActor.remove();
                    }

                    handCards.clearChildren();
                    cardStack.clearChildren();
                }),
                new StartRoundOverlay(this, round)
        ));
    }

    public void startTrick() {
        clearActivePlayer();
        execute(() -> cardStack.clearChildren());
    }

    /**
     * Indicates that the next call to {@link #setHand(UUID, List)} should be animated.
     */
    public void setJuggling(boolean juggling) {
        execute(() -> {
            this.juggling = juggling;
            if (juggling) {
                handCards.setSelectMode(CardsGroup.SelectMode.SINGLE);
            } else {
                handCards.setSelectMode(CardsGroup.SelectMode.NONE);
                juggledCard = null;
                handCards.setSelected(null);
            }
        });
    }

    public void finishTrick(UUID player, List<Card> cards) {
        var seat = seats.get(player);

        var action = parallel();
        execute(sequence(
                run(() -> cardStack.removeAll().forEach(card -> {
                    var angle = (card.getRotation() % 360 + 360) % 360;
                    var rotation = rotateTo(angle < 90 || angle > 270 ? 0 : 180, AnimationTimings.STACK_FINISH_ROTATE);
                    rotation.setUseShortestDirection(true);

                    action.addAction(sequence(
                            targeting(card, changeParent(game.data.stage.getRoot())),
                            parallel(
                                    targeting(card, rotation),
                                    targeting(card, moveTo(
                                            seat.getFrontX() - card.getWidth() / 2,
                                            seat.getFrontY() - card.getHeight() / 2,
                                            AnimationTimings.STACK_FINISH_MOVE
                                    ))
                            ),
                            targeting(card, alpha(0, AnimationTimings.STACK_FINISH_FADE)),
                            removeActor(card)
                    ));
                })),
                action
        ));
    }

    public void setHand(UUID player, List<Card> cards) {
        if (isSelf(player)) {
            var sequence = sequence();
            sequence.addAction(run(() -> {
                var changes = handCards.update(cards);

                // animate card changes
                if (juggling) {
                    setJuggling(false);
                    sequence.addAction(animateJuggle(changes.first(), changes.second()));
                }
            }));
            execute(sequence);
        }
    }

    public void setTrump(Card trumpCard, Card.Suit trumpSuit) {
        if (trumpCardActor == null) {
            trumpCardActor = new CardActor(Card.HIDDEN, atlas);
        }

        if (trumpSuitActor == null) {
            trumpSuitActor = new CardActor(Card.HIDDEN, atlas);
        }

        String player = null;
        if (activePlayer != null && activePlayer.second() == PICK_TRUMP) {
            player = state.getPlayers().get(activePlayer.first());
            clearActivePlayer();
        }

        execute(new TrumpOverlay(this, player, trumpCard, trumpSuit));
    }

    public void addPrediction(int round, UUID player, int prediction) {
        boolean changed = false;

        if (activePlayer != null && activePlayer.first().equals(player) && (activePlayer.second() == CHANGE_PREDICTION || activePlayer.second() == MAKE_PREDICTION)) {
            changed = activePlayer.second() == CHANGE_PREDICTION;
            clearActivePlayer();
        }

        if (isSelf(player)) {
            addMessage(game.messages.format("game.action." + (changed ? "change" : "make") + "_prediction.self", prediction));
        } else {
            var name = state.getPlayers().get(player);
            addMessage(game.messages.format("game.action."  + (changed ? "change" : "make") + "_prediction.other", name, prediction));
        }

        execute(() -> padOfTruth.setPrediction(players.indexOf(player), round, prediction));
    }

    public void playCard(UUID player, Card card) {
        if (activePlayer != null && activePlayer.first().equals(player) && activePlayer.second() == PLAY_CARD) {
            clearActivePlayer();
        }

        if (isSelf(player)) {
            addMessage(game.messages.get("game.action.play_card.self"));
        } else {
            var name = state.getPlayers().get(player);
            addMessage(game.messages.format("game.action.play_card.other", name));
        }

        Seat seat = seats.get(player);

        var sequence = sequence();
        sequence.addAction(run(() -> {
            CardActor actor = null;
            if (isSelf(player)) {
                actor = handCards.remove(card);
                actor.setOrigin(actor.getWidth() / 2, actor.getHeight() / 2);
            }

            if (actor == null) {
                actor = seat.createHandCard(card, atlas);
            }

            cardStack.add(seat, actor);
            sequence.addAction(delay(actor));
            sequence.addAction(delay(AnimationTimings.STACK_HOLD));
        }));
        execute(sequence);
    }

    public void addScores(int round, Map<UUID, Integer> scores) {
        execute(() -> {
            for (int i = 0; i < players.size(); i++) {
                UUID player = players.get(i);
                padOfTruth.setScore(i, round, scores.get(player));
            }
        });
    }

    public void clearActivePlayer() {
        setActivePlayer(null, null, 0);
    }

    public void setActivePlayer(UUID player, UserInputMessage.Action action, long timeout) {
        if (action == SYNC) throw new IllegalArgumentException();

        // reset label color
        if (activePlayer != null && nameLabels.containsKey(activePlayer.first())) {
            var label = nameLabels.get(activePlayer.first());
            execute(() -> label.setStyle(labelStyleDefault));
        }

        if (player == null && action == null) {
            activePlayer = null;
            setPersistentMessage(null);
            return;
        }

        activePlayer = Triple.of(player, action, timeout);
        // set label color
        if (nameLabels.containsKey(player)) {
            var label = nameLabels.get(player);
            execute(() -> label.setStyle(labelStyleActive));
        }

        boolean isSelf = state.getSelf().equals(player);
        if (isSelf || player == null) {
            // show interface
            setPersistentMessage(null);
            switch (action) {
                case PICK_TRUMP -> execute(new PickTrumpOverlay(this, timeout, false));
                case MAKE_PREDICTION -> execute(new MakePredictionOverlay(this, timeout, state.getRound()));
                case CHANGE_PREDICTION -> execute(new ChangePredictionOverlay(this, timeout, state.getRound(), state.getPredictions().get(state.getSelf())));
                case PLAY_CARD -> setPersistentMessage(game.messages.get("game.message.play_card.self"));
            }
        }

        if (!isSelf) {
            // show message
            var key = switch (action) {
                case CHANGE_PREDICTION -> "game.message.change_prediction.";
                case JUGGLE_CARD -> "game.message.juggle_card.";
                case MAKE_PREDICTION -> "game.message.make_prediction.";
                case PICK_TRUMP -> "game.message.pick_trump.";
                case PLAY_CARD -> "game.message.play_card.";
                default -> throw new AssertionError();
            };

            if (player == null) {
                setPersistentMessage(game.messages.get(key + "all"));
            } else {
                var name = state.getPlayers().get(player);
                setPersistentMessage(game.messages.format(key + "other", name));
            }
        }
    }

    public void addMessage(@Nls String text) {
        addMessage(text, false);
    }

    public void addMessage(@Nls String text, boolean immediate) {
        var label = new Label(text, game.data.skin);
        label.addAction(sequence(
                delay(AnimationTimings.MESSAGE_HOLD),
                alpha(0, AnimationTimings.MESSAGE_FADE),
                removeActor()
        ));

        Runnable runnable = () -> {
            if (persistentMessage != null && persistentMessage.getParent() == messages) {
                messages.addActorBefore(persistentMessage, label);
            } else {
                messages.addActor(label);
            }
        };

        if (immediate) {
            runnable.run();
        } else {
            execute(runnable);
        }
    }

    public void setPersistentMessage(@Nls String text) {
        execute(() -> {
            if (persistentMessage != null) {
                persistentMessage.addAction(sequence(
                        delay(AnimationTimings.MESSAGE_HOLD),
                        alpha(0, AnimationTimings.MESSAGE_FADE),
                        removeActor()
                ));
                persistentMessage = null;
            }

            if (text != null) {
                persistentMessage = new Label(text, getData().skin);
                messages.addActor(persistentMessage);
            }
        });
    }

    public void send(PlayerMessage message) {
        if (!sending.getAndSet(true) || (message instanceof ContinueMessage)) {
            game.getClient().send(new InteractionMessage(message));
        } else {
            addMessage("Please slow down.", true);
        }
    }

    public void ready(boolean success) {
        if (!pendingSync.get()) {
            sending.set(false);
        }

        if (success && currentAction instanceof Overlay overlay && overlay instanceof InteractionOverlay) {
            overlay.finish();
        }

        if (checkAction(JUGGLE_CARD) && juggledCard != null) {
            handCards.setSelected(juggledCard);
            juggledCard = null;
        }
    }

    public void timeout() {
        addMessage(game.messages.get("game.message.timeout"));
        ready(true);
        clearActivePlayer();
    }

    public void sync() {
        pendingSync.set(true);
        sending.set(true);
    }

    @Override
    public void dispose() {
        if (atlas != null) atlas.dispose();
    }

    public WizardGame.Data getData() {
        return game.data;
    }

    public I18NBundle getMessages() {
        return game.messages;
    }

    private Action animateJuggle(List<CardActor> removed, List<CardActor> added) {
        // find left- and rightmost seat
        Seat tmpLeft = null, tmpRight = null;
        for (Seat seat : seats.values()) {
            if (seat == Seat.BOTTOM) continue;
            if (tmpLeft == null || tmpLeft.compareTo(seat) > 0) tmpLeft = seat;
            if (tmpRight == null || tmpRight.compareTo(seat) < 0) tmpRight = seat;
        }
        if (tmpLeft == null) return null;
        var left = tmpLeft;
        var right = tmpRight;

        var animation = parallel();
        removed.forEach(actor -> {
            getData().stage.addActor(actor);
            animation.addAction(targeting(actor, left.moveToHand(AnimationTimings.JUGGLE)));
        });

        added.forEach(actor -> {
            // capture old values taking consideration of the animated layout changes in CardsGroup
            float x = actor.getX(), y = actor.getY();
            var actions = actor.getActions();
            for (int i = 0; i < actions.size; i ++) {
                var action = actions.get(i);
                if (action instanceof MoveToAction move) {
                    x = move.getX();
                    y = move.getY();
                    actions.removeIndex(i--);
                }
            }
            var rotation = actor.getRotation();

            // apply start values
            var startPos = new Vector2(right.getHandX(), right.getHandY());
            handCards.stageToLocalCoordinates(startPos);
            actor.setOrigin(actor.getWidth() / 2, actor.getHeight() / 2);
            animation.addAction(targeting(actor, moveTo(startPos.x, startPos.y)));
            animation.addAction(targeting(actor, rotateTo(right.getHandAngle())));

            // animate
            animation.addAction(targeting(actor, moveTo(x, y, AnimationTimings.JUGGLE)));
            animation.addAction(targeting(actor, rotateTo(rotation, AnimationTimings.JUGGLE)));
        });

        var cleanup = parallel();
        removed.forEach(actor -> cleanup.addAction(removeActor(actor)));

        return sequence(animation, cleanup);
    }

    @Getter
    public enum Seat {
        BOTTOM(WizardGame.WIDTH * 0.5f, 0, 0, 0, Align.bottom, WizardGame.WIDTH * 0.5f, 300),
        LEFT(0, WizardGame.HEIGHT * 0.5f, 50, WizardGame.HEIGHT * 0.5f + 110f, Align.bottomLeft, 117.5f, WizardGame.HEIGHT * 0.5f),
        TOP_LEFT(WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT - 200),
        TOP(WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT - 200),
        TOP_RIGHT(WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT - 200),
        RIGHT(WizardGame.WIDTH, WizardGame.HEIGHT * 0.5f, WizardGame.WIDTH - 50, WizardGame.HEIGHT * 0.5f + 110f, Align.bottomRight, WizardGame.WIDTH - 117.5f, WizardGame.HEIGHT * 0.5f);

        // position of the hand, should be offscreen
        private final float handX;
        private final float handY;
        private final float handAngle;

        // position of the label
        private final float labelX;
        private final float labelY;
        private final int labelAlign;

        private final float frontX;
        private final float frontY;

        Seat(float handX, float handY, float labelX, float labelY, int labelAlign, float frontX, float frontY) {
            this.handX = handX;
            this.handY = handY;
            this.labelX = labelX;
            this.labelY = labelY;
            this.labelAlign = labelAlign;
            this.frontX = frontX;
            this.frontY = frontY;

            var deltaX = WizardGame.WIDTH * 0.5f - handX;
            var deltaY = WizardGame.HEIGHT * 0.5f - handY;
            this.handAngle = (float) Math.toDegrees(Math.atan2(deltaY, deltaX) + Math.PI / 2);
        }

        public CardActor createHandCard(Card card, TextureAtlas atlas) {
            var actor = new CardActor(card, atlas);
            actor.setPosition(getHandX() - actor.getWidth() / 2, getHandY() - actor.getHeight() / 2);
            actor.setRotation(getHandAngle());
            actor.setOrigin(actor.getWidth() / 2, actor.getHeight() / 2);
            return actor;
        }

        public Action moveToHand(float duration) {
            return parallel(
                    moveTo(handX, handY, duration),
                    rotateTo(handAngle, duration)
            );
        }
    }
}
